import contextlib
from typing import Iterator

try:
    from RPi import GPIO
except ModuleNotFoundError:
    from qbee_gpio.gpio.types import GPIO


@contextlib.contextmanager
def gpio_setup(mode: int = GPIO.BCM) -> Iterator[None]:
    """Setup and cleanup GPIO through context manager."""
    GPIO.setmode(mode)
    try:
        yield None
    finally:
        GPIO.cleanup()



import asyncio
import sys

from qbee_gpio.config import QbeeConfig
from qbee_gpio.orchestrator import QbeeOrchestrator


def run() -> None:
    if "--init-config" in sys.argv:
        QbeeConfig.load().save()
        return
    asyncio.run(QbeeOrchestrator(debug="-v" in sys.argv).run())



import asyncio
import base64
import contextlib
import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import AsyncIterator, Optional

logger = logging.getLogger(__name__)


@dataclass
class NowPlaying:
    artist: str
    album: str
    title: str

    def display(self, lines: int) -> str:
        if lines == 1:
            return self.title
        elif lines == 2:
            return f"{self.artist}\n{self.title}"
        else:
            return f"{self.artist}\n{self.album}\n{self.title}"


class ShairportNowPlayingPoller(contextlib.AsyncExitStack, asyncio.Protocol):
    """Poll shairport metadata pipe and expose minimal information."""

    RE_METADATA = re.compile(
        # Album info
        f"<item><type>{b'core'.hex()}</type><code>{b'asal'.hex()}</code><length>[0-9]+</length>\n"
        '<data encoding="base64">\n(.*?)</data></item>(?:.|\n)*?'
        # Artist info
        f"<item><type>{b'core'.hex()}</type><code>{b'asar'.hex()}</code><length>[0-9]+</length>\n"
        '<data encoding="base64">\n(.*?)</data></item>(?:.|\n)*?'
        # Title info
        f"<item><type>{b'core'.hex()}</type><code>{b'minm'.hex()}</code><length>[0-9]+</length>\n"
        '<data encoding="base64">\n(.*?)</data></item>'
    )

    def __init__(self, path: Path):
        super().__init__()
        self.path = path
        # We need a queue as `data_received` is not async.
        self._received_data: asyncio.Queue[bytes] = asyncio.Queue()

    async def poll(self) -> AsyncIterator[NowPlaying]:
        async with self:
            logger.debug("started now playing poller")
            last_event: Optional[NowPlaying] = None
            async for data in self._receive():
                if match := self.RE_METADATA.search(data.decode()):
                    album, artist, title = tuple(
                        base64.b64decode(e).decode() for e in match.groups()
                    )
                    event = NowPlaying(
                        artist=artist,
                        album=album,
                        title=title,
                    )
                    if event == last_event:
                        continue
                    logger.debug("now playing %r", event)
                    last_event = event
                    yield event

    async def __aenter__(self):
        pipe = self.enter_context(
            open(
                os.open(str(self.path), os.O_RDONLY),
            ),
        )
        transport, _ = await asyncio.get_running_loop().connect_read_pipe(
            lambda: self,
            pipe,
        )
        self.callback(transport.close)
        return self

    def data_received(self, data: bytes) -> None:
        self._received_data.put_nowait(data)

    async def _receive(self) -> AsyncIterator[bytes]:
        while True:
            data = await self._received_data.get()
            yield data



import logging
from pathlib import Path
from typing import ClassVar, Literal, Optional

import attrs
import zenconfig


@attrs.define
class LCDConfig:
    enable: bool = True
    width: int = 16
    lines: Literal[1, 2, 4] = 2
    # GPIO PIN configuration (BCM mode).
    pin_power: int = 17
    pin_register_select: int = 27
    pin_enable: int = 18
    pin_data_4: int = 25
    pin_data_5: int = 24
    pin_data_6: int = 23
    pin_data_7: int = 22
    now_playing_path: Path = attrs.field(
        default=Path("/tmp/shairport-sync-metadata"),
        converter=Path,
    )
    startup_message: str = "Qbee"


@attrs.define
class SoundDetectionConfig:
    enable: bool = True
    driver_path: Path = attrs.field(
        default=Path("/dev/snd/pcmC0D0p"),
        converter=Path,
    )
    # Number of seconds to keep amp on after sound has stopped.
    standby_duration: Optional[float] = 600
    # Change this as needed or set to None to disable shutdown after standby.
    # The default command will work if
    #   - `$_USER_ $_HOST_ = (root) NOPASWD: /usr/sbin/shutdown` has been added to sudoers
    #   - or the user is root
    shutdown_command: Optional[str] = "sudo shutdown -h now"
    # GPIO PIN configuration (BCM mode).
    pin_amp_power: int = 4


@attrs.define
class QbeeConfig(zenconfig.Config):
    PATH: ClassVar[str] = "~/.qbee.yaml"

    sound_detection: SoundDetectionConfig = SoundDetectionConfig()
    lcd: LCDConfig = LCDConfig()

    # Logging.
    log_config: dict = attrs.field(
        factory=lambda: {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "formatter": {
                    "validate": True,
                    "format": "%(asctime)s %(levelname)-5.5s [%(name)s] %()messages",
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "formatter",
                    "stream": "ext://sys.stdout",
                },
            },
            "root": {
                "level": logging.getLevelName(logging.INFO),
                "handlers": ["console"],
            },
        },
    )
